@@ -24,6 +24,7 @@ fn fangs(
     n_sweet: Rval,
     a: Rval,
     n_cores: Rval,
+    use_neighbors: Rval,
     quiet: Rval,
 ) -> Rval {
     let mut timer = EchoTimer::new();
@@ -42,6 +43,7 @@ fn fangs(
         .num_threads(n_cores)
         .build()
         .unwrap();
+    let use_neighbors = use_neighbors.as_bool();
     let quiet = quiet.as_bool();
     let status_file = match std::env::var("FANGS_STATUS") {
         Ok(x) => Path::new(x.as_str()).to_owned(),
@@ -163,8 +165,9 @@ fn fangs(
         if interrupted || r::check_user_interrupt() {
             panic!("Caught user interrupt before main loop, so aborting.");
         }
-        let loss = expected_loss_from_samples(z.view(), &views, a, &pool);
-        initials.push((z, loss, rng));
+        let weight_matrices = make_weight_matrices(z.view(), &views, a, &pool);
+        let loss = expected_loss_from_weight_matrices(&weight_matrices[..], &pool);
+        initials.push((z, loss, weight_matrices, rng));
     }
     if timer.echo() {
         interrupted |= rprint!(
@@ -182,11 +185,10 @@ fn fangs(
         initials
             .into_par_iter()
             .enumerate()
-            .map(|(id, (z, loss, rng))| {
-                let weight_matrices = make_weight_matrices(z.view(), &views, a, &pool);
+            .map(|(id, (z, loss, weight_matrices, rng))| {
                 let n_accepts = 0;
                 let when = 1;
-                (z, weight_matrices, loss, id, n_accepts, when, rng)
+                (z, loss, weight_matrices, id, n_accepts, when, rng)
             })
             .collect()
     });
@@ -203,64 +205,89 @@ fn fangs(
     let seconds_in_initialization = timer.total_as_secs_f64();
     let mut period_timer = PeriodicTimer::new(1.0);
     let mut iteration_counter = 0;
-    while iteration_counter < n_iterations && timer.total_as_secs_f64() < max_seconds {
-        iteration_counter += 1;
+    if use_neighbors {
         pool.install(|| {
-            sweets.par_iter_mut().for_each(
-                |(z, weight_matrices, loss, _, n_accepts, when, rng)| {
-                    let n_features = z.ncols();
-                    let total_length = n_items * n_features;
-                    fn index_1d_to_2d(index: usize, ncols: usize) -> [usize; 2] {
-                        [index / ncols, index % ncols]
-                    }
-                    let index = index_1d_to_2d(rng.gen_range(0..total_length), n_features);
-                    flip_bit(z, weight_matrices, a, index, &views);
-                    let new_loss = expected_loss_from_weight_matrices(&weight_matrices, &pool);
-                    if new_loss < *loss {
-                        *n_accepts += 1;
-                        *when = iteration_counter;
-                        *loss = new_loss;
-                    } else {
-                        flip_bit(z, weight_matrices, a, index, &views);
-                    }
-                },
-            );
+            sweets
+                .par_iter_mut()
+                .for_each(|(z, loss, weight_matrices, _, _, _, _)| {
+                    *loss = neighborhood_sweeten(
+                        z,
+                        &mut weight_matrices[..],
+                        &views[..],
+                        n_items,
+                        a,
+                        &pool,
+                        max_seconds,
+                        &timer,
+                    );
+                })
         });
-        if !quiet || status_file.exists() {
-            period_timer.maybe(iteration_counter == n_iterations, || {
-                if quiet && status_file.exists() {
+    } else {
+        let n_iterations = if n_iterations == 0 {
+            sweets
+                .iter()
+                .map(|x| x.0.nrows() * x.0.ncols())
+                .max()
+                .unwrap_or(0)
+        } else {
+            n_iterations
+        };
+        while iteration_counter < n_iterations && timer.total_as_secs_f64() < max_seconds {
+            iteration_counter += 1;
+            pool.install(|| {
+                sweets.par_iter_mut().for_each(
+                    |(z, loss, weight_matrices, _, n_accepts, when, rng)| {
+                        let n_features = z.ncols();
+                        let total_length = n_items * n_features;
+                        let index = index_1d_to_2d(rng.gen_range(0..total_length), n_features);
+                        flip_bit(z, weight_matrices, a, index, &views);
+                        let new_loss = expected_loss_from_weight_matrices(&weight_matrices, &pool);
+                        if new_loss < *loss {
+                            *n_accepts += 1;
+                            *when = iteration_counter;
+                            *loss = new_loss;
+                        } else {
+                            flip_bit(z, weight_matrices, a, index, &views);
+                        }
+                    },
+                );
+            });
+            if !quiet || status_file.exists() {
+                period_timer.maybe(iteration_counter == n_iterations, || {
+                    if quiet && status_file.exists() {
+                        interrupted |= rprint!(
+                            "{}",
+                            format!(
+                                "*** {} exists, so forcing status display.\n",
+                                status_file.display()
+                            )
+                            .as_str()
+                        );
+                        r::flush_console();
+                    }
+                    sweets.sort_unstable_by(|x, y| x.1.partial_cmp(&y.1).unwrap());
+                    let best = sweets.first().unwrap();
                     interrupted |= rprint!(
                         "{}",
                         format!(
-                            "*** {} exists, so forcing status display.\n",
-                            status_file.display()
-                        )
-                        .as_str()
-                    );
-                    r::flush_console();
-                }
-                sweets.sort_unstable_by(|x, y| x.2.partial_cmp(&y.2).unwrap());
-                let best = sweets.first().unwrap();
-                interrupted |= rprint!(
-                    "{}",
-                    format!(
                         "\rIter. {}: Since iter. {}, E(loss) is {:.4} from #{} with {} accept{}.",
                         iteration_counter,
                         best.5,
-                        best.2,
+                        best.1,
                         best.3 + 1,
                         best.4,
                         if best.4 == 1 { "" } else { "s" }
                     )
-                    .as_str()
-                );
+                        .as_str()
+                    );
+                    r::flush_console();
+                });
+            }
+            if interrupted || r::check_user_interrupt() {
+                rprint!("\nCaught user interrupt, so breaking out early.");
                 r::flush_console();
-            });
-        }
-        if interrupted || r::check_user_interrupt() {
-            rprint!("\nCaught user interrupt, so breaking out early.");
-            r::flush_console();
-            break;
+                break;
+            }
         }
     }
     if !quiet {
@@ -278,8 +305,8 @@ fn fangs(
         );
         r::flush_console();
     }
-    sweets.sort_unstable_by(|x, y| x.2.partial_cmp(&y.2).unwrap());
-    let (best_z, _, best_loss, sweeten_number, n_accepts, best_iteration, _) =
+    sweets.sort_unstable_by(|x, y| x.1.partial_cmp(&y.1).unwrap());
+    let (best_z, best_loss, _, sweeten_number, n_accepts, best_iteration, _) =
         sweets.swap_remove(0);
     if timer.echo() {
         rprint!(
@@ -340,6 +367,111 @@ fn fangs(
     list
 }
 
+#[roxido]
+fn fangs_double_greedy(samples: Rval, max_seconds: Rval, a: Rval, n_cores: Rval) -> Rval {
+    let timer = EchoTimer::new();
+    let o = samples.get_list_element(0);
+    if !o.is_double() || !o.is_matrix() {
+        panic!("All elements of 'samples' must be double matrices.");
+    }
+    let n_items = o.nrow();
+    let n_samples = samples.len();
+    if n_samples < 1 {
+        panic!("Number of samples must be at least one.");
+    }
+    let max_seconds = max_seconds.as_f64();
+    let a = a.as_f64();
+    let n_cores = n_cores.as_usize();
+    let pool = rayon::ThreadPoolBuilder::new()
+        .num_threads(n_cores)
+        .build()
+        .unwrap();
+    let mut max_n_features_observed = 0;
+    let mut views = Vec::with_capacity(n_samples);
+    for i in 0..n_samples {
+        let o = samples.get_list_element(i);
+        if !o.is_double() || !o.is_matrix() || o.nrow() != n_items {
+            panic!("All elements of 'samples' must be double matrices with a consistent number of rows.");
+        }
+        let view = make_view(o);
+        max_n_features_observed = max_n_features_observed.max(view.ncols());
+        views.push(view)
+    }
+    let mut z = Array2::<f64>::zeros((n_items, max_n_features_observed));
+    let mut weight_matrices = make_weight_matrices(z.view(), &views[..], a, &pool);
+    let loss = neighborhood_sweeten(
+        &mut z,
+        &mut weight_matrices[..],
+        &views[..],
+        n_items,
+        a,
+        &pool,
+        max_seconds,
+        &timer,
+    );
+    let (estimate, estimate_slice) = Rval::new_matrix_double(n_items, z.ncols(), pc);
+    let mut index = 0;
+    for j in 0..z.ncols() {
+        for i in 0..n_items {
+            estimate_slice[index] = z[(i, j)];
+            index += 1;
+        }
+    }
+    let list = Rval::new_list(3, pc);
+    list.names_gets(rval!(["estimate", "expectedLoss", "secondsTotal",]));
+    list.set_list_element(0, estimate);
+    list.set_list_element(1, rval!(loss));
+    list.set_list_element(2, rval!(timer.total_as_secs_f64()));
+    list
+}
+
+fn neighborhood_sweeten(
+    z: &mut Array2<f64>,
+    weight_matrices: &mut [Array2<f64>],
+    views: &[ArrayView2<f64>],
+    n_items: usize,
+    a: f64,
+    pool: &ThreadPool,
+    max_seconds: f64,
+    timer: &EchoTimer,
+) -> f64 {
+    let mut outer_loss = expected_loss_from_weight_matrices(&weight_matrices[..], pool);
+    loop {
+        if timer.echo() {
+            println!("Current loss: {}", outer_loss);
+        }
+        if timer.total_as_secs_f64() >= max_seconds {
+            break;
+        }
+        // Optimize within a given number of columns
+        let mut best_candidate_loss = f64::INFINITY;
+        let mut best_index = [0, 0];
+        for i in 0..n_items {
+            for j in 0..z.ncols() {
+                let candidate_loss = expected_loss_from_weight_matrices_if_flip_bit(
+                    &z,
+                    weight_matrices,
+                    a,
+                    [i, j],
+                    &views,
+                    &pool,
+                );
+                if candidate_loss < best_candidate_loss {
+                    best_index = [i, j];
+                    best_candidate_loss = candidate_loss;
+                }
+            }
+        }
+        if best_candidate_loss < outer_loss {
+            flip_bit(z, weight_matrices, a, best_index, &views);
+            outer_loss = best_candidate_loss;
+        } else {
+            break;
+        }
+    }
+    outer_loss
+}
+
 #[roxido]
 fn fangs_old(
     samples: Rval,
@@ -504,9 +636,6 @@ fn fangs_old(
                 .for_each(|(z, weight_matrices, loss, _, n_accepts, when, rng)| {
                     let n_features = z.ncols();
                     let total_length = n_items * n_features;
-                    fn index_1d_to_2d(index: usize, ncols: usize) -> [usize; 2] {
-                        [index / ncols, index % ncols]
-                    }
                     let index = index_1d_to_2d(rng.gen_range(0..total_length), n_features);
                     flip_bit(z, weight_matrices, a, index, &views);
                     let new_loss = expected_loss_from_weight_matrices(&weight_matrices, &pool);
@@ -761,10 +890,14 @@ fn make_weight_matrices(
     })
 }
 
+fn index_1d_to_2d(index: usize, ncols: usize) -> [usize; 2] {
+    [index / ncols, index % ncols]
+}
+
 #[allow(clippy::float_cmp)]
 fn flip_bit(
     z: &mut Array2<f64>,
-    matrices: &mut Vec<Array2<f64>>,
+    matrices: &mut [Array2<f64>],
     a: f64,
     index: [usize; 2],
     samples: &[ArrayView2<f64>],
@@ -772,9 +905,8 @@ fn flip_bit(
     let old_bit = z[index];
     z[index] = if old_bit == 0.0 { 1.0 } else { 0.0 };
     let b = 2.0 - a;
-    let i0 = index[0];
-    let i1 = index[1];
-    let result = samples.iter().zip(matrices.iter_mut()).for_each(|(zz, w)| {
+    let [i0, i1] = index;
+    samples.iter().zip(matrices.iter_mut()).for_each(|(zz, w)| {
         for i2 in 0..w.ncols() {
             let bit_in_sample = if i2 >= zz.ncols() { 0.0 } else { zz[[i0, i2]] };
             w[[i1, i2]] += if old_bit == 0.0 {
@@ -804,7 +936,63 @@ fn flip_bit(
         max_ulps = 4
     );
     */
-    result
+}
+
+fn update_w(
+    zz: &ArrayView2<f64>,
+    w: &mut Array2<f64>,
+    i0: usize,
+    i1: usize,
+    a: f64,
+    b: f64,
+    bit: f64,
+) {
+    for i2 in 0..w.ncols() {
+        let bit_in_sample = if i2 >= zz.ncols() { 0.0 } else { zz[[i0, i2]] };
+        w[[i1, i2]] += if bit == 0.0 {
+            if bit_in_sample == 0.0 {
+                a
+            } else {
+                -b
+            }
+        } else {
+            if bit_in_sample == 0.0 {
+                -a
+            } else {
+                b
+            }
+        };
+    }
+}
+
+#[allow(clippy::float_cmp)]
+fn expected_loss_from_weight_matrices_if_flip_bit(
+    z: &Array2<f64>,
+    matrices: &mut [Array2<f64>],
+    a: f64,
+    index: [usize; 2],
+    samples: &[ArrayView2<f64>],
+    pool: &ThreadPool,
+) -> f64 {
+    let old_bit = z[index];
+    let new_bit = if old_bit == 0.0 { 1.0 } else { 0.0 };
+    let b = 2.0 - a;
+    let [i0, i1] = index;
+    pool.install(|| {
+        samples
+            .par_iter()
+            .zip(matrices.par_iter_mut())
+            .fold(
+                || 0.0,
+                |acc: f64, (zz, w)| {
+                    update_w(zz, w, i0, i1, a, b, old_bit);
+                    let lss = loss(w);
+                    update_w(zz, w, i0, i1, a, b, new_bit);
+                    acc + lss
+                },
+            )
+            .reduce(|| 0.0, |a, b| a + b)
+    }) / (matrices.len() as f64)
 }
 
 #[allow(clippy::float_cmp)]
